static unsigned socket_packet_in_buffers(unsigned socket, unsigned long ack_num)
{
  unsigned n = OUT_PACKETS;

  while(n--)
    if(outgoing_packets[n].socket == socket
       && outgoing_packets[n].ack_num == ack_num) break;
  return n;
}

static unsigned socket_packets_in_buffers(unsigned socket)
{
  unsigned retval = 0;
  unsigned n = OUT_PACKETS;

  while(n--)
    if(outgoing_packets[n].socket == socket)
      retval++;
  return retval;
}

static int too_much_ploss(unsigned socket)
{
  return (sockets[socket]->ploss > OUT_PACKETS-socket_packets_in_buffers(socket));
}

static void add_outpacket(unsigned socket, unsigned char __far *packet, unsigned len, unsigned payload_size)
{
  unsigned n = OUT_PACKETS;
  unsigned long ack_num = sockets[socket]->send.count + sockets[socket]->send.iss
                     + sockets[socket]->send.ack_waiting;

  if(socket_packet_in_buffers(socket, ack_num) != -1) return;
  if(too_much_ploss(socket)) return;
  if(!((sockets[socket]->out_buff_read_offset + sockets[socket]->send.ack_waiting)
     - sockets[socket]->out_buff_write_offset)
     && sockets[socket]->last_frame_size < (SOCKET_OUT_BUFF_SIZE>>1)) return;

  while(n--)
  {
    if(outgoing_packets[n].socket == -1)
    {
      outgoing_packets[n].socket = socket;
      if(len < 60)
      {
        _fmemset(&outgoing_packets[n].packet_buff[len], 0, 60-len);
        len = 60;
      }
      outgoing_packets[n].length = len;
      outgoing_packets[n].id = sockets[socket]->send.psh_count++;
      outgoing_packets[n].payload_size = payload_size;
      sockets[socket]->send.count += sockets[socket]->send.ack_waiting;
      outgoing_packets[n].ack_num = ack_num;
      outgoing_packets[n].last_sent = 0;
      sockets[socket]->out_buff_read_offset += sockets[socket]->send.ack_waiting;
      sockets[socket]->status = 1;
      _fmemcpy(outgoing_packets[n].packet_buff, packet, len);
      break;
    }
  }
  sockets[socket]->send.retries = 0;
}

static unsigned resend_outpacket4(unsigned socket)
{
  unsigned n;
  unsigned id;
  unsigned p;
  unsigned long time;
  unsigned count;
  unsigned char o;

  if(sockets[socket]->send.retries >= MAX_RETRIES)
    return 0;
  else
  {
    time = ttime(0);
    for(n=OUT_PACKETS;n--;)
      if(outgoing_packets[n].socket == socket
         && time >= outgoing_packets[n].last_sent + TCP_RETRANSMISSION_TIME)
      {
        id = outgoing_packets[n].id;
        sockets[socket]->ploss++;
        break;
      }
    if(n == -1) return 0;
    else
    {
      sockets[socket]->time = time;
      do
        for(n=OUT_PACKETS;n--;)
          if(outgoing_packets[n].socket == socket
             && outgoing_packets[n].id == id - 1)
          {
            id = outgoing_packets[n].id;
            break;
          }
      while(n != -1);

      for(count=OUT_PACKETS;count--;o=2)
      {
        for(p=id,n=OUT_PACKETS;n--;)
        {
          if(outgoing_packets[n].socket == socket)
            if(outgoing_packets[n].id == id)
            {
              if(send_raw_packet(outgoing_packets[n].packet_buff, outgoing_packets[n].length))
                goto end;
              outgoing_packets[n].last_sent = time;
              o--;
            }
            else if(outgoing_packets[n].id == id+1)
            {
              p++;
              o--;
            }
          if(!o) break;
        }
        if(p != ++id)
          break;
      }
    }
    end:
    sockets[socket]->send.retries++;
    return 1;
  }
}

static void mark_outpacket_sent(unsigned index)
{
  unsigned socket = outgoing_packets[index].socket;
  unsigned id = outgoing_packets[index].id;
  unsigned p;
  unsigned retries = sockets[socket]->send.retries;

  outgoing_packets[index].socket = -1;

  do
    for(id--,p=OUT_PACKETS;p--;)
      if(outgoing_packets[p].socket == socket
        && outgoing_packets[p].id == id)
      {
        outgoing_packets[p].socket = -1;
        break;
      }
  while(p != -1);

  if(retries > 1)
  {
    if(!resend_outpacket4(socket))
      tcp4_packetizer(socket);
  }
  else
    sockets[socket]->time = ttime(0);
}

static void socket_delete_buffers(unsigned socket)
{
  unsigned n = OUT_PACKETS;

  while(n--)
    if(outgoing_packets[n].socket == socket)
      outgoing_packets[n].socket = -1;
}

static int out_buffers_free(void)
{
  unsigned n = OUT_PACKETS;

  while(n--)
    if(outgoing_packets[n].socket == -1)
      break;

  return n+1;
}
