static int get_dns_request_by_port(unsigned port)
{
  int n = DNS_CACHE_SIZE;

  while(n--)
    if(dns_cache[n].local_port == port)
      break;

  return n;
}

static unsigned short __far *udp4_pseudoheader(struct ipv4_addr src_addr, struct ipv4_addr dst_addr, unsigned len);
static unsigned udp4_checksum(unsigned char __far *packet, unsigned short __far *pseudoheader, unsigned len);

static int sender_mac_changed(unsigned socket, unsigned char *sender_mac, struct ipv4_addr src_addr)
{
  unsigned m;
  int a;

  for(m=6;m-- && !sockets[socket]->remote_addr[m];);
  if(m == -1)
    _fmemcpy(sockets[socket]->remote_addr, sender_mac, 6);
  else if(_fmemcmp(sockets[socket]->remote_addr, sender_mac, 6))
  {
    #ifdef ARP
    if((a = address_in_arp_cache4(src_addr)) >= 0)
    {
      if(_fmemcmp(arp_cache[a].mac, sender_mac, 6))
        return 1;
      else
        _fmemcpy(sockets[socket]->remote_addr, sender_mac, 6);        
    }      
    else
    #endif
      return 1;
  }
  return 0;
}

static void parse_udp4(unsigned char __far *packet, struct udp4_header *header, struct udp_data *tcp4_data, unsigned packet_len, struct ipv4_addr src_addr, struct ipv4_addr dst_addr, unsigned char *sender_mac)
{
  unsigned i, e;
  int a;
  unsigned short __far *shortptr = (unsigned short __far*)packet;
  unsigned m;
  
  header->src_port = switch_endianness_s(shortptr[0]);
  header->dst_port = switch_endianness_s(shortptr[1]);
  header->len = switch_endianness_s(shortptr[2]);

  if(packet_len < header->len)
  {
    return;
  }

  if(src_addr.p1 != 127 &&
    udp4_checksum(packet, udp4_pseudoheader(src_addr, dst_addr, packet_len), packet_len))
    {
      #ifdef DROP_BAD_CHECKSUM_UDP_PACKETS
        rx_error_count++;
        return;
      #endif
    }

  // if the packet is a response to DNS query
  if(!memcmp(&src_addr, &ipv4_dns, sizeof(struct ipv4_addr)) && (i = get_dns_request_by_port(header->dst_port)) != -1
    && !parse_dns_ans(&packet[8], &dns_cache[i], packet_len-8));
  else if((i = find_socket_by_ports_and_remote_addr4(header->dst_port, header->src_port, src_addr, 1)) == -1)
  {
    if((e = find_open_port_by_port(header->dst_port, 1)) != -1)
    {
      // if port is open and this seems to be a new connection, create new socket
      if((i = find_free_socket_index()) == -1)
        return;
      if(push_socket_queue(e, sockets[i]->socket_descriptor))
        delete_socket(i, 1);
      else
      {
        sockets[i]->ip_addr.ipv4_addr = src_addr;
        sockets[i]->remote_port = header->src_port;
        sockets[i]->local_port = header->dst_port;
        sockets[i]->status = sockets[i]->udp = 1;
        sockets[i]->time = ttime(0);
        goto recv;
      }
      return;                              
    }
    else if(header->dst_port == 68 && header->src_port == 67)
      parse_dhcp_reply(&packet[8], packet_len-8);
  }
  else
  {
    if(sender_mac_changed(i, sender_mac, src_addr))
      return; // sender's physical address changed, drop packet
    recv:udp4_recv_socket(i, &packet[8], packet_len-8);
  }
}

static unsigned udp4_checksum(unsigned char __far *packet, unsigned short __far *pseudoheader, unsigned len)
{
  register unsigned short checksum = 0;
  register unsigned short addition;
  unsigned hlen;
  unsigned short __far *spacket = (unsigned short __far*)packet;

  if(len & 1)
  {
    len--;
    checksum = packet[len];
  }

  hlen = len>>1;

  while(hlen--)
  {
    addition = spacket[hlen];
    checksum += addition;
    if(checksum < addition)
      if(!++checksum)
        checksum++;
  }

  len = 6;

  while(len--)
  {
    addition = pseudoheader[len];
    checksum += addition;
    if(checksum < addition)
      if(!++checksum)
        checksum++;
  }

//  if(checksum == 0xFFFF)
//    checksum = ~checksum;

  return ~checksum;
}

static void build_udp4_header(unsigned char __far *packet, struct udp4_header *header)
{
  unsigned short __far *shortptr = (unsigned short __far*)packet;

  shortptr[0] = switch_endianness_s(header->src_port);
  shortptr[1] = switch_endianness_s(header->dst_port);
  shortptr[2] = switch_endianness_s(header->len);
  shortptr[3] = 0;
}

static unsigned short __far *udp4_pseudoheader(struct ipv4_addr src_addr, struct ipv4_addr dst_addr, unsigned len)
{
  static char __based (__segname("_CODE")) retval[12];
  unsigned short __far *lenptr = (unsigned short __far*)&retval[10];

  retval[0] = src_addr.p1;
  retval[1] = src_addr.p2;
  retval[2] = src_addr.p3;
  retval[3] = src_addr.p4;

  retval[4] = dst_addr.p1;
  retval[5] = dst_addr.p2;
  retval[6] = dst_addr.p3;
  retval[7] = dst_addr.p4;

  retval[8] = 0;
  retval[9] = 17;
  *lenptr = switch_endianness_s(len);
  
  
  return (unsigned short*)retval;
}

static unsigned build_udp4_packet(unsigned char __far *packet, struct udp4_header *header, struct udp_data *udp_data, struct ipv4_addr *dst_addr)
{
  unsigned short __far *checksum = (unsigned short __far*)&packet[6];
  unsigned char __far *udp_payload = &packet[8];
  unsigned short len = udp_data->len;

  _fmemcpy(udp_payload, udp_data->data, len);

 
  build_udp4_header(packet, header);
  *checksum = udp4_checksum(packet, udp4_pseudoheader(ipv4_my_addr, *dst_addr, 8+udp_data->len), 8+udp_data->len);

  return 8 + udp_data->len;
}

static int send_udp4(struct ipv4_addr dst, struct udp4_header *udp_header, struct udp_data *udp_data)
{
  unsigned total_length;
  unsigned ipv4_header_length;
  struct ipv4_header ipv4_header;  
  unsigned char __far *packet;
  unsigned char __far *payload;
  int n;

  packet = get_free_async_buff();
  payload = &packet[14];

  init_ipv4_packet_header(dst, packet, 0);
  
  ipv4_header.type_of_service = 0;
  ipv4_header.identification = random();
  ipv4_header.flags.dont_fragment = 0;
  ipv4_header.flags.more_fragments = 0;
  ipv4_header.fragment_offset = 0;
  ipv4_header.time_to_live = 0xFF;
  ipv4_header.protocol = 17;

  ipv4_header.src_address = ipv4_my_addr;
  ipv4_header.dst_address = dst;

  ipv4_header_length = build_ipv4_header(payload, &ipv4_header);

  udp_header->len = udp_data->len + 8;

  total_length = ipv4_header_length + build_udp4_packet(&payload[ipv4_header_length], udp_header, udp_data, &dst);

  ipv4_inject_lengthinfo(payload, total_length);
  ipv4_inject_header_checksum(payload, ipv4_header_length);


  return send_ip_packet(packet, total_length+14);
 
}
